//----------------------------------------------------------------
//	ATtiny2313-based dual serial motor controller
//	Martin Neil Weiss
//	9/27/09
// 
//  rev.1: corrections to if-then statements in setMotorSpeed
//----------------------------------------------------------------
#define F_CPU 14745600

#include <avr/io.h>
#include <util/delay.h> 
#include <avr/interrupt.h>

#define LEDport 		PORTD
#define redLED   		(1<<0)			//PORTD0
#define greenLED 		(1<<1)			//PORTD1
#define USART_BAUDRATE 	28800
#define BAUD_PRESCALE 	(((F_CPU / (USART_BAUDRATE * 16UL))) - 1)
#define RxBufferSize	32
#define RxBufferMask	(RxBufferSize - 1)

char RxBuffer[RxBufferSize];
volatile char RxBuffer_head=0, RxBuffer_tail=0;

//macros...
#define abs(x) 					((x)>=0?(x):(-x))
#define sgn(x)             		((x)==0?0:(x)>0?1:-1) 
//-----------------------------------------------------------------
ISR(_VECTOR(7)) //vector(7) = USART RX vector
{
	unsigned char RxByte, tmphead;

//store received byte...
	RxByte = UDR;

//calculate index to store it in in the ring buffer...
	tmphead = (RxBuffer_head + 1) & RxBufferMask;
	RxBuffer_head = tmphead;

//store in ring buffer...
	if (tmphead != RxBuffer_tail){
		RxBuffer[tmphead] = RxByte;
	}
}
//-----------------------------------------------------------------
void clearRxBuffer(){

//fill the ring buffer with zeros, set head = tail = 0...
	unsigned char i;

	for(i=0;i<RxBufferSize;i++)
		RxBuffer[i] = '0';

	RxBuffer_head = RxBuffer_tail = 0;

}
//-----------------------------------------------------------------
char readRxBuffer(){

//read one character from the ring buffer...
	char tmptail;

//wait until there's a character in the buffer...
	while(RxBuffer_head == RxBuffer_tail);

//calculate the index of the ring buffer to pull the data from...
	tmptail = (RxBuffer_tail + 1) & RxBufferMask;
	RxBuffer_tail = tmptail;

//return the characer at the tail of the ring buffer...
	return RxBuffer[tmptail];
}
//-----------------------------------------------------------------
unsigned char strlen(char *s){

unsigned char i;

	for(i=0;*s!='\0';s++){
		i++;
	}
	return i;
}
//-----------------------------------------------------------------
void serout_string(char temp[]){

unsigned char i;

	for(i=0;i<=strlen(temp)-1;i++){
		while((UCSRA & (1<<UDRE))==0){};
		UDR = temp[i];
	}
}
//-----------------------------------------------------------------
void serout_int(int word){

char str[6];
char i, j;
int digit, val;

	i = 0;
	val = abs(word);
	do{
		digit = val;
		val /= 10;
		digit -= val * 10;
		str[i++] = '0' + digit;
	}while( val != 0 );

	if(word < 0)
		str[i++]='-';

	for(j=i-1;j>=0;j--){
		while((UCSRA & (1<<UDRE))==0){};
		UDR = str[j];
	}
}
//-----------------------------------------------------------------
void delay_ms( unsigned int time_ms ){

//delay for a user-specified number of milliseconds...
	unsigned int i;

	for(i = 0; i < time_ms; i++){
		_delay_ms( 1 );
	}
}
//-----------------------------------------------------------------
int computePWM(long int speed){

const long int deadband = 75;
const long int minSpeed = 15; 
long int temp;

	if(abs(speed) < minSpeed)
		temp = 0;	
	else{
		temp = ((speed * (255 - deadband)) >> 8) + sgn(speed) * deadband;
	
		if(abs(temp) > 255)
			temp = sgn(temp) * 255;
	}

	return temp;
}
//-----------------------------------------------------------------
void setMotorSpeed(unsigned char ID, char command){

//set the speed of either motor 1 or 2
//note each call to this function sets the speed of only 1 motor
//also note we use 2 PWM channels per motor (less components on the board)

static int motor1speed=0, motor2speed=0;
int pwm, newSpeed=0;
char accelFactor = 25;

	if (ID == '1'){

	//compute new speed...
		switch(command){
			case '>': newSpeed = motor1speed + accelFactor;break;
			case '<': newSpeed = motor1speed - accelFactor;break;
			case '0': newSpeed = 0;break;
		}
		
		if (abs(newSpeed) > 255)
			newSpeed = sgn(newSpeed) * 255;

	//stop motor1 for 1 ms if reversing direction...
		if (sgn(motor1speed) != sgn(newSpeed)){
			OCR0A = OCR0B = 0;
			delay_ms(1);
		}

	//compute pwm needed for new speed...
		motor1speed = newSpeed;
		pwm = computePWM(motor1speed);

	//set motor1 speed...
		if (pwm >= 0){
			OCR0A = pwm;
			OCR0B = 0;
		}else{
			OCR0A = 0;
			OCR0B = abs(pwm);
		}

	}else if (ID == '2'){

	//compute new speed...
		switch(command){
			case '>': newSpeed = motor2speed + accelFactor;break;
			case '<': newSpeed = motor2speed - accelFactor;break;
			case '0': newSpeed = 0;break;
		}
		
		if (abs(newSpeed) > 255)
			newSpeed = sgn(newSpeed) * 255;

	//stop motor2 for 1 ms if reversing direction...
		if (sgn(motor2speed) != sgn(newSpeed)){
			OCR1A = OCR1B = 0;
			delay_ms(1);
		}

	//compute pwm needed for new speed...
		motor2speed = newSpeed;
		pwm = computePWM(motor2speed);

	//set motor2 speed...
		if (pwm >= 0){
			OCR1A = pwm;
			OCR1B = 0;
		}else{
			OCR1A = 0;
			OCR1B = abs(pwm);
		}
	}

	//return a status character to indiate a healthy packet...
	//printf_P(PSTR("Motor %d speed = %d\r\n"), ID-'0', newSpeed);
	serout_string(" Motor ");
	serout_int(ID-'0');
	serout_string(" speed = ");
	serout_int(newSpeed);
}
//-----------------------------------------------------------------
void initialize(){

//disable all interrupts
	cli(); 

// setup ports as input / output (1 = output, 0 = input)...
    DDRB 	= 0b00011100;
	DDRD 	= 0b00100010;

// setup timer0 for 8 bit fast PWM at 7.2 kHz
// 14745600 Hz / 256 / 8 = 7200 Hz
	TCCR0B 	= 0;			//stop timer0
	OCR0A  	= 0;			//set duty cycle to zero
	OCR0B	= 0;			//set duty cycle to zero
	TCCR0A 	= 0b10100011;	//clear on compare match, fast PWM
	TCCR0B 	= 0b00000010;	//TOP = 255, pre-scalar = 8	

// setup timer1 for 8 bit fast PWM at 7.2 kHz
// 14745600 Hz / 256 / 8 = 7200 Hz
	TCCR1B 	= 0;			//stop timer1
	OCR1A  	= 0;			//set duty cycle to zero
	OCR1B	= 0;			//set duty cycle to zero
	TCCR1A 	= 0b10100001;	//clear on compare match, fast 8 bit PWM
	TCCR1B 	= 0b00001010;	//TOP = 255, pre-scalar = 8	

// disable interrupts...
	GIMSK	= 0;
	TIMSK	= 0;
	PCMSK	= 0;

// USART...
	UCSRB |= (1<<RXEN)|(1<<TXEN);	// enable receive + transmit
	UCSRC |= (1<<UCSZ0)|(1<<UCSZ1);	//8-bit characters, 1 stop bit, no parity
	UBRRL  = BAUD_PRESCALE;			// load low byte into UBRR register
	UBRRH  = (BAUD_PRESCALE>>8);	// load high byte into UBRR register
	UCSRB |= (1<<RXCIE);			// enable USRAT interrupt

//re-enable interrupts
	sei(); 
}
//-----------------------------------------------------------------
int main(void){

const int packetSize = 4;
unsigned char RxPacket[packetSize], i;

//We will use the USART to receive "packets" that indicate how to set the speed
//  of the each of the two motors.
//Packet format = {(1/2)(</0/>)}, where (1/2) indicates motor 1 or 2 
//  and (</>) indicates acceleration or decceleration. 
//The character sequence {1>} causes motor 1 to accelerate.
//The character sequence {1<} causes motor 1 to deccelerate.
//The character sequence {10} causes motor 1 to stop.
//The character sequence {2>} causes motor 2 to accelerate.
//The character sequence {2<} causes motor 2 to deccelerate.
//The character sequence {20} causes motor 2 to stop.

//set up microcontroller...
	initialize();
	
//initialize USART receive buffer...
	clearRxBuffer();

//main loop...
	for(;;){

	//wait until we receive a "start-of-packet" character...
		do{
			RxPacket[0] = readRxBuffer();
		}while(RxPacket[0] != '{');

	//read the rest of the packet...
		for(i=1;i<=packetSize-1;i++){
			RxPacket[i] = readRxBuffer();
		}

		if (RxPacket[3] == '}'){
			
			//motorID = RxPacket[1];
			
			//if (RxPacket[2] == '<')
			//	inc = -1;
			//else if (RxPacket[2] == '>')
			//	inc = 1;
			//else
			//	inc = 0;

			//setMotorSpeed(motorID - '0', inc*accelFactor);
			setMotorSpeed(RxPacket[1],RxPacket[2]);
		}
	}
}
//-----------------------------------------------------------------
